/*	Direct3D OpenGL driver
 *
 * Copyright 2001-2002 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "winbase.h"
#include "debugtools.h"
#include "ddrawi.h"
#include "d3d8.h"
#include "d3dhal.h"
#include "d3dgl_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define GL_SetupViewport(x)

static inline void
GL_draw_vertices(GL_D3D_priv* priv, GLenum gltype, DWORD fvf,
		 DWORD start, DWORD count,
		 GL_IndexFunc index, LPVOID index_data,
		 DWORD ibase)
{
    if (priv->d3dp->renderstate[D3DRENDERSTATE_LIGHTING] && (fvf & D3DFVF_NORMAL || priv->d3d_version >= 8))
	GL_draw_unlit_vertices(priv, gltype, fvf, start, count, index, index_data, ibase);
    else assert(FALSE);
}

static DWORD GL_DirectIndex(DWORD i, LPVOID p)
{
    return i;
}

static DWORD GL_IndexedIndex(DWORD i, LPVOID p)
{
    return ((LPWORD)p)[i];
}

static GLfloat* accum_feedback_rect(LPD3D_P d3dp, GLfloat *feedback, LPRECT rect, int v)
{
    int x1, y1, x2, y2;
    x1 = floorf(feedback[0]);
    x2 = ceilf(feedback[0]);
#ifndef HAVE_OSMESA
    y1 = d3dp->surf->lpGbl->wHeight - (int)ceilf(feedback[1]);
    y2 = d3dp->surf->lpGbl->wHeight - (int)floorf(feedback[1]);
#else
    y1 = floorf(feedback[1]);
    y2 = ceilf(feedback[1]);
#endif
    /* TRACE("  (%d,%d) <= (%f,%f)\n", x1, y1, feedback[0], feedback[1]); */
    if (!v) {
	rect->left = x1;
	rect->right = x2;
	rect->top = y1;
	rect->bottom = y2;
    } else {
	if (x1 < rect->left) rect->left = x1;
	if (x2 > rect->right) rect->right = x2;
	if (y1 < rect->top) rect->top = y1;
	if (y2 > rect->bottom) rect->bottom = y2;
    }
    return feedback+2;
}

static int calc_feedback_rect(GL_D3D_priv* priv, GLfloat *feedback, int count, LPRECT rect)
{
    LPD3D_P d3dp = priv->d3dp;
    int v = 0;
    GLfloat *end = feedback + count;
    while (feedback < end) {
	int n, tok = *feedback++;
	switch (tok) {
	case GL_POINT_TOKEN:
	    /* TRACE(" POINT\n"); */
	    feedback = accum_feedback_rect(d3dp, feedback, rect, v++);
	    break;
	case GL_LINE_TOKEN:
	case GL_LINE_RESET_TOKEN:
	    /* TRACE(" LINE\n"); */
	    feedback = accum_feedback_rect(d3dp, feedback, rect, v++);
	    feedback = accum_feedback_rect(d3dp, feedback, rect, v++);
	    break;
	case GL_POLYGON_TOKEN:
	    n = *feedback++;
	    /* TRACE(" POLYGON(%d)\n", n); */
	    while (n--)
		feedback = accum_feedback_rect(d3dp, feedback, rect, v++);
	    break;
	default:
	    FIXME("unhandled feedback token %d\n", tok);
	}
    }
    return v;
}

void GL_GetPrimitiveType(LPD3DHAL_DP2COMMAND cmd,
				GLenum* lppt, LPDWORD lpdwCount,
				LPBOOL lpbIndexed)
{
    DWORD dwCount;
    GLenum pt;
    BOOL indexed = FALSE;

    switch (cmd->bCommand)
    {
    case D3DDP2OP_POINTS:
	/* Special case. */
	TRACE("Draw POINTS\n");
	pt = GL_POINTS;
	dwCount = 0;
	indexed = FALSE;
	break;

    case D3DDP2OP_INDEXEDLINELIST2:
	indexed = TRUE;
    case D3DDP2OP_LINELIST:
	TRACE("Draw LINES\n");
	pt = GL_LINES;
	dwCount = cmd->u.wPrimitiveCount*2;
	break;

    case D3DDP2OP_INDEXEDLINESTRIP:
	indexed = TRUE;
    case D3DDP2OP_LINESTRIP:
	TRACE("Draw LINE_STRIP\n");
	pt = GL_LINE_STRIP;
	dwCount = cmd->u.wPrimitiveCount+1;
	break;

    case D3DDP2OP_INDEXEDTRIANGLELIST2:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLELIST:
	TRACE("Draw TRIANGLES\n");
	pt = GL_TRIANGLES;
	dwCount = cmd->u.wPrimitiveCount*3;
	break;

    case D3DDP2OP_INDEXEDTRIANGLESTRIP:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLESTRIP:
	TRACE("Draw TRIANGLE_STRIP\n");
	pt = GL_TRIANGLE_STRIP;
	dwCount = cmd->u.wPrimitiveCount+2;
	break;

    case D3DDP2OP_INDEXEDTRIANGLEFAN:
	indexed = TRUE;
    case D3DDP2OP_TRIANGLEFAN:
	TRACE("Draw TRIANGLE_FAN\n");
	pt = GL_TRIANGLE_FAN;
	dwCount = cmd->u.wPrimitiveCount+2;
	break;

    default:
	ERR("Unhandled primitive\n");
	pt = 0;
	dwCount = 0;
	indexed = FALSE;
	break;
    }

    *lpdwCount = dwCount;
    *lppt = pt;
    *lpbIndexed = indexed;
}

HRESULT D3D_GL_draw(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    const WORD* indices;
    DWORD fvf;
    DWORD dwCount;
    DWORD dwIndexBase;
    DWORD dwVStart, dwVCount;
    DWORD flags = priv->glFlags;
    GLenum pt;
    BOOL indexed;
    BOOL calc_rect = FALSE, invert = FALSE;
    GLfloat *feedback = NULL;

    GL_GetPrimitiveType(cmd, &pt, &dwCount, &indexed);

    indices = indexed ? data2 : NULL;
    dwIndexBase = indexed ? ((LPD3DHAL_DP2STARTVERTEX)data1)->wVStart : 0;
    dwVStart = priv->dp2_dwVertexOffset;
    dwVCount = priv->dp2_dwVertexLength;

    if (priv->callbacks.D3dRenderLock)
      calc_rect = priv->callbacks.D3dRenderLock(priv->d3dp->dwhContext, TRUE, &invert);

    if (invert) priv->glFlags |= GLF_RENDER_INVERTED;
    else priv->glFlags &= ~GLF_RENDER_INVERTED;

    if ((flags ^ priv->glFlags) & GLF_RENDER_INVERTED) {
      TRACE("invert=%d\n", invert);
      priv->glFlags |= GLF_NEW_TRANSFORM;
#ifndef HAVE_OSMESA
      glFrontFace(invert ? GL_CCW : GL_CW);
#else
      glFrontFace(invert ? GL_CW : GL_CCW);
#endif
    }

    D3D_GL_update_palettes(priv);

    /* Assume we can until proven otherwise. */
    priv->glFlags |= GLF_USE_VERTEX_ARRAYS;

    if (priv->glFlags & GLF_VERTEX_PROGRAM) {
	fvf = 0;
	/* load up vertex program stuff */
	D3D_GL_setup_programmable(priv, data1, indices, pt, dwCount, indexed);
	if (calc_rect && priv->callbacks.D3dRenderArea)
	  priv->callbacks.D3dRenderArea(priv->d3dp->dwhContext, NULL);
    } else {
	fvf = D3D_GL_setup_fixed(priv, data1, indices, pt, dwCount, indexed);
	if (!fvf) {
	  if (priv->callbacks.D3dRenderUnlock)
	    priv->callbacks.D3dRenderUnlock(priv->d3dp->dwhContext, FALSE);
	  return E_FAIL;
	}
	if (calc_rect && priv->callbacks.D3dRenderArea) {
	  unsigned mp = (pt == GL_POINTS) ? 3 : 14;
	  unsigned size = mp*cmd->u.wPrimitiveCount;
	  TRACE("entering feedback mode to compute bounds (buffer size %d)\n", size);
	  feedback = HeapAlloc(GetProcessHeap(), 0, size*sizeof(GLfloat));
	  glFeedbackBuffer(size, GL_2D, feedback);
	  glRenderMode(GL_FEEDBACK);
	}
    }

render_start:
    TRACE("start\n");

    if (pt == GL_POINTS)
    {
	LPD3DHAL_DP2POINTS dat = (LPD3DHAL_DP2POINTS)data2;
	DWORD p;

	for (p=0; p<cmd->u.wPrimitiveCount; p++,dat++) {
	    if (priv->glFlags & GLF_USE_VERTEX_ARRAYS)
		glDrawArrays(pt, dat->wVStart, dat->wCount);
	    else
		GL_draw_vertices(priv, pt, fvf, dat->wVStart, dat->wCount,
				 GL_DirectIndex, NULL, dwIndexBase);
	}
    }
    else if (indexed)
    {
	if (priv->glFlags & GLF_USE_VERTEX_ARRAYS) {
#if 1
	    glDrawRangeElements(pt, dwVStart - dwIndexBase,
				dwVStart - dwIndexBase + dwVCount,
				dwCount, GL_UNSIGNED_SHORT, indices);
#else
	    glDrawElements(pt, dwCount, GL_UNSIGNED_SHORT, indices);
#endif
	} else
	    GL_draw_vertices(priv, pt, fvf, 0, dwCount,
			     GL_IndexedIndex, data2, dwIndexBase);
    }
    else
    {
	LPD3DHAL_DP2STARTVERTEX dat = (LPD3DHAL_DP2STARTVERTEX)data1;

	if (priv->glFlags & GLF_USE_VERTEX_ARRAYS)
	    glDrawArrays(pt, dwVStart + dat->wVStart, dwCount);
	else
	    GL_draw_vertices(priv, pt, fvf, dwVStart + dat->wVStart, dwCount,
			     GL_DirectIndex, NULL, dwIndexBase);
    }

    CHECK_GL_ERROR("DrawPrimitive");

    if (feedback) {
	int count = glRenderMode(GL_RENDER);
	RECT rect;
	TRACE("left feedback mode, %d entries\n", count);
	if (calc_feedback_rect(priv, feedback, count, &rect))
	  priv->callbacks.D3dRenderArea(priv->d3dp->dwhContext, &rect);
	HeapFree(GetProcessHeap(), 0, feedback);
	feedback = NULL;
	/* now do the actual rendering */
	goto render_start;
    }

    if (priv->glFlags & GLF_VERTEX_PROGRAM) {
        /* finish up vertex program stuff */
        D3D_GL_finish_programmable(priv);
    }
    else {
	D3D_GL_finish_fixed(priv);
    }

    if (priv->callbacks.D3dRenderUnlock)
      priv->callbacks.D3dRenderUnlock(priv->d3dp->dwhContext, TRUE);

    TRACE("done\n");
    return D3D_OK;
}

HRESULT D3D_GL_clear(GL_D3D_priv* priv, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2CLEAR dat = (LPD3DHAL_DP2CLEAR)data1;
    GLboolean ztest = GL_FALSE;
    GLbitfield mask = 0;

    /* For the moment, ignore the rectangles */

    TRACE("Flags=%8lx\n", dat->dwFlags);

    GL_SetupViewport(This);

    if (dat->dwFlags & D3DCLEAR_TARGET) {
	mask |= GL_COLOR_BUFFER_BIT;
	glClearColor(RGBA_GETRED(dat->dwFillColor) / 255.0,
		     RGBA_GETGREEN(dat->dwFillColor) / 255.0,
		     RGBA_GETBLUE(dat->dwFillColor) / 255.0,
		     RGBA_GETALPHA(dat->dwFillColor) / 255.0);
    }
    if (dat->dwFlags & D3DCLEAR_STENCIL) {
	mask |= GL_STENCIL_BUFFER_BIT;
	glClearStencil(dat->dwFillStencil);
    }
    if (dat->dwFlags & D3DCLEAR_ZBUFFER) {
	mask |= GL_DEPTH_BUFFER_BIT;
	glClearDepth(dat->dvFillDepth);
	glGetBooleanv(GL_DEPTH_WRITEMASK, &ztest);
	glDepthMask(GL_TRUE); /* Enables Z writing to be sure to delete also the Z buffer */
    }

    if (priv->callbacks.D3dRenderLock) {
      BOOL invert = FALSE;
      if (priv->callbacks.D3dRenderLock(priv->d3dp->dwhContext, TRUE, &invert)) {
        RECT rect;
        /* same rectangle as D3D_GL_viewport_info works with */
        rect.left = priv->d3dp->viewport.dwX;
        rect.right = rect.left + priv->d3dp->viewport.dwWidth;
        rect.top = priv->d3dp->viewport.dwY;
        rect.bottom = rect.top + priv->d3dp->viewport.dwHeight;
        if ((dat->dwFlags & D3DCLEAR_TARGET) &&
            (rect.right > rect.left) && (rect.bottom > rect.top))
          priv->callbacks.D3dRenderArea(priv->d3dp->dwhContext, &rect);
      }
    }

    glClear(mask);

    if (priv->callbacks.D3dRenderUnlock)
      priv->callbacks.D3dRenderUnlock(priv->d3dp->dwhContext, TRUE);

    if (dat->dwFlags & D3DCLEAR_ZBUFFER)
	glDepthMask(ztest);

    return D3D_OK;
}
